from hubmap_api_py_client.internal import InternalClient

_default_limit = 1000


class ExternalClient():
    def __init__(self, base_url):
        self.client = InternalClient(base_url)

    def __repr__(self):
        return f'<Client base_url={self.client.base_url}>'

    def _query(
            self,
            input_type=None, output_type=None, has=None,
            genomic_modality=None, limit=None, p_value=None,
            ResultsSetSubclass=None):
        if not isinstance(has, list):
            raise TypeError(f'"has" parameter must be a list, not {has}')
        handle = self.client.hubmap_query(
            input_type, output_type, has,
            genomic_modality, limit, p_value)
        return ResultsSetSubclass(
            self.client, handle,
            input_type=input_type, output_type=output_type,
            query=has
        )


def _add_method(output_type, ResultsSetSubclass):
    method_name = f'select_{output_type}s'
    method = (
        lambda self, where=None, has=None,
        genomic_modality=None, limit=_default_limit, p_value=None:
        self._query(
            input_type=where, output_type=output_type, has=has,
            genomic_modality=genomic_modality, limit=limit, p_value=p_value,
            ResultsSetSubclass=ResultsSetSubclass)
    )
    setattr(ExternalClient, method_name, method)


class ResultsSet():
    def __init__(
            self, client, handle,
            input_type=None, output_type=None,
            query=None):
        self.client = client
        self.handle = handle
        self.input_type = input_type
        self.output_type = output_type
        self.query = query

    def __repr__(self):
        return (
            f'<{class_name(self.output_type)} '
            f'base_url={self.client.base_url} handle={self.handle}>')

    def __len__(self):
        return self.client.set_count(self.handle, self.output_type)

    def __or__(self, other_set):
        return self._operation(other_set, self.client.set_union)

    def __and__(self, other_set):
        return self._operation(other_set, self.client.set_intersection)

    def __sub__(self, other_set):
        return self._operation(other_set, self.client.set_difference)

    def _operation(self, other_set, method):
        if self.output_type != other_set.output_type:
            raise ValueError(
                'Operand output types do not match: '
                f'{self.output_type} != {other_set.output_type}')
        new_handle = method(self.handle, other_set.handle, self.output_type)
        return ResultsSet(
            self.client, new_handle,
            input_type=self.input_type, output_type=self.output_type,
            query=self.query)

    def __getitem__(self, key):
        if isinstance(key, int):
            if key < 0:
                raise ValueError('Negative index not supported')
            return self._get_list(1, offset=key)[0]
        if isinstance(key, slice):
            if key.step is not None:
                raise ValueError('Step is not supported')
            if key.start is None or key.stop is None:
                raise ValueError('Start and stop are required')
            if key.start < 0 or key.stop < 0:
                raise ValueError('Start and stop must be >= 0')
            if key.stop < key.start:
                raise ValueError('Stop must be > start')
            return self._get_list(key.stop - key.start, offset=key.start)
        raise TypeError()

    def _get_list(self, limit, offset=0):
        return self.client.set_list_evaluation(self.handle, self.output_type, limit, offset=offset)

    def get_details(
            self, limit, offset=0,
            values_included=[], sort_by=None):
        return self.client.set_detail_evaluation(
            self.handle, self.output_type, limit,
            offset=offset,
            sort_by=sort_by,
            values_type=self.input_type,
            values_included=[self.query])


def class_name(output_type):
    return f'{output_type.capitalize()}ResultsSet'


def _create_subclass(output_type):
    return type(class_name(output_type), (ResultsSet,), {})


for output_type in ['cell', 'organ', 'gene', 'cluster']:
    ResultsSetSubclass = _create_subclass(output_type)
    _add_method(output_type, ResultsSetSubclass)



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

with open("VERSION", "r") as fh:
    version = fh.read()

setuptools.setup(
    name="hubmap-api-py-client",
    version=version,
    install_requires=[
        # Keep in sync with requirements-lower-bound.txt:
        'requests>=2.0.0'
    ],
    # scripts=[],
    author="Chuck McCallum",
    author_email="mccallucc+cells_client@gmail.com",
    description="Client for the HuBMAP Cells API",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/hubmapconsortium/hubmap-api-py-client",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    # Keep in sync with .travis.yml:
    python_requires='>=3.6',
    # f-strings aren't available in 3.5.
)



from typing import List

import requests


HANDLE = 'query_pickle_hash'


class InternalClient():
    def __init__(self, base_url):
        # 'https://cells.dev.hubmapconsortium.org/api/'
        self.base_url = base_url

    def _check_parameters(
            self,
            input_type: str, output_type: str, input_set: List[str],
            genomic_modality: str, p_value: float):
        output_types = ['cell', 'organ', 'gene', 'cluster']
        if output_type not in output_types:
            raise ValueError(f'{output_type} not in {output_types}')

        input_types = {
            # Allowed input types vary depending on output type
            'cell': ['gene', 'organ', 'protein', 'dataset'],
            'organ': ['cell', 'gene'],
            'gene': ['organ', 'cluster'],
            'cluster': ['gene']
        }
        if input_type not in input_types[output_type]:
            raise ValueError(f'{input_type} not in {input_types[output_type]}')

        genomic_modalities = ['rna', 'atac']  # Used for quantitative gene->cell queries
        if input_type == 'gene' and output_type == 'cell':
            if genomic_modality not in genomic_modalities:
                raise ValueError(f'{genomic_modality} not in {genomic_modalities}')

        if (input_type == 'organ' and output_type == 'gene'
                or input_type == 'gene' and output_type == 'organ'):
            if p_value is None or p_value < 0 or p_value > 1:
                raise ValueError(f'p_value {p_value} should be in [0,1]')

    def _fill_request_dict(
            self,
            input_type: str, output_type: str, input_set: List[str],
            genomic_modality: str, p_value: float):
        request_dict = {'input_type': input_type, 'input_set': input_set}
        if input_type == 'gene' and output_type == 'cell':
            request_dict['genomic_modality'] = genomic_modality
        if input_type in ['organ', 'gene'] and output_type == 'gene':
            request_dict['p_value'] = p_value
        if genomic_modality is not None:
            request_dict['genomic_modality'] = genomic_modality
        request_dict['logical_operator'] = "and"
        return request_dict

    def hubmap_query(
            self,
            input_type: str, output_type: str, input_set: List[str],
            genomic_modality: str = None, limit: int = 1000, p_value: float = None):
        '''
        This function takes query parameters and returns a query set token.
        '''
        self._check_parameters(
            input_type=input_type, output_type=output_type,
            input_set=input_set, genomic_modality=genomic_modality, p_value=p_value)
        request_url = self.base_url + output_type + "/"
        request_dict = self._fill_request_dict(
            input_type, output_type, input_set, genomic_modality, p_value)
        response = requests.post(request_url, request_dict)
        results = response.json()['results']
        # Returns the key to be used in future computations
        return results[0][HANDLE]

    # These functions take two query set tokens and return an API token:

    def set_intersection(
            self, set_key_one: str, set_key_two: str, set_type: str) -> str:
        return self._operation(set_key_one, set_key_two, set_type, 'intersection/')

    def set_union(
            self, set_key_one: str, set_key_two: str, set_type: str) -> str:
        return self._operation(set_key_one, set_key_two, set_type, 'union/')

    def set_difference(
            self, set_key_one: str, set_key_two: str, set_type: str) -> str:
        return self._operation(set_key_one, set_key_two, set_type, 'difference/')

    def _operation(
            self, set_key_one: str, set_key_two: str, set_type: str, path: str) -> str:
        request_url = self.base_url + path
        request_dict = {"key_one": set_key_one, "key_two": set_key_two, "set_type": set_type}
        results = requests.post(request_url, request_dict).json()['results']
        # Returns the key to be used in future computations
        return results[0][HANDLE]

    # These functions take a query set token and return an evaluated query_set:

    def _check_detail_parameters(
            self, set_type, values_type):
        type_map = {'cell': ['gene', 'protein'], 'gene': [
            'organ', 'cluster'], 'cluster': ['gene'], 'organ': ['gene']}
        allowed_types = type_map[set_type]
        if values_type not in allowed_types:
            raise ValueError(
                f'For "{set_type}", only {allowed_types} allowed, not "{values_type}"')

    def set_count(
            self, set_key: str, set_type: str) -> str:
        request_url = self.base_url + "count/"
        request_dict = {"key": set_key, "set_type": set_type}
        results = requests.post(request_url, request_dict).json()['results']
        return results[0]["count"]

    def set_list_evaluation(
            self, set_key: str, set_type: str, limit: int, offset: int = 0):
        '''
        This function/API call returns a minimal version of the set,
        containing a list of cells/genes/etc w/o
        associated quantitative values.  It should be reasonably fast.
        '''
        request_url = self.base_url + set_type + "evaluation/"
        request_dict = {"key": set_key, "set_type": set_type, "limit": limit}
        response = requests.post(request_url, request_dict)
        results = response.json()['results']
        return results  # Returns the key to be used in future computations

    def set_detail_evaluation(
            self, set_key: str, set_type: str, limit: int,
            values_included: List = [], sort_by: str = None, values_type: str = None,
            offset: int = 0):
        '''
        This function/API call returns a more detailed version of the set,
        containing data specified in include_values
        It may be slow.
        '''
        self._check_detail_parameters(set_type, values_type)
        request_url = self.base_url + set_type + "detailevaluation/"
        request_dict = {"key": set_key, "set_type": set_type, "limit": limit,
                        "values_included": values_included, "sort_by": sort_by,
                        "values_type": values_type}
        response = requests.post(request_url, request_dict)
        results = response.json()['results']
        return results  # Returns the key to be used in future computations
